@@ -1,7 +1,7 @@
 # Namespace dance baby
-try:
+try:  # pragma: no cover
     import pkg_resources
     pkg_resources.declare_namespace(__name__)
-except ImportError:
+except ImportError:  # pragma: no cover
     import pkgutil
     __path__ = pkgutil.extend_path(__path__, __name__)
