@@ -13,6 +13,7 @@ Packages: http://repo.reverbrain.com/
 
 """
 
+import os
 import types
 
 import itertools
@@ -50,6 +51,11 @@ DEFAULT_GROUPS = [1]
 class Storage(driver.Base):
 
     def __init__(self, path=None, config=None):
+        # Turn on streaming support
+        self.supports_bytes_range = True
+        # Increase buffer size up to 640 Kb
+        self.buffer_size = 128 * 1024
+        # Create default Elliptics config
         cfg = elliptics.Config()
         # The parameter which sets the time to wait for the operation complete
         cfg.config.wait_timeout = int(config.get("elliptics_wait_timeout",
@@ -74,7 +80,7 @@ class Storage(driver.Base):
             self.groups = map(int, self.groups.strip('[]').split(','))
 
         if len(self.groups) == 0:
-            raise ValueError("Specify groups")
+            raise exceptions.ConfigError("elliptics_groups must be specified")
 
         # loglevel of elliptics logger
         elliptics_log_level = int(config.get('elliptics_verbosity', 0))
@@ -89,26 +95,28 @@ class Storage(driver.Base):
 
         remotes_configuration = config.get('elliptics_nodes')
         if remotes_configuration is None:
-            raise ValueError("elliptics_nodes must be specified")
+            raise exceptions.ConfigError("elliptics_nodes must be specified")
         elif isinstance(remotes_configuration,
                         (types.TupleType, types.ListType)):
             remotes = remotes_configuration
         elif isinstance(remotes_configuration, types.StringTypes):
             remotes = remotes_configuration.split()
         else:
-            raise ValueError("elliptics_nodes must be list, tuple or string")
+            raise exceptions.ConfigError("elliptics_nodes must be list,"
+                                         "tuple or string")
 
-        at_least_one = False
         for remote in remotes:
             try:
-                logger.debug("Add remote %s", remote)
+                logger.debug("Remote %s is being added", remote)
                 self._elliptics_node.add_remote(remote)
-                at_least_one = True
+                logger.info("%s remote has been added successfully", remote)
             except Exception as err:
                 logger.error("Failed to add remote %s: %s", remote, err)
 
-        if not at_least_one:
-            raise Exception("Unable to connect to Elliptics")
+        if not self._session.routes.addresses():
+            # routing table is empty,
+            # as no remotes have been successfully added or conencted.
+            raise exceptions.ConnectionError("Unable to connect to Elliptics")
 
     @property
     def _session(self):
@@ -141,13 +149,13 @@ class Storage(driver.Base):
         r.wait()
         err = r.error()
         if err.code != 0:
-            logger.warning("Unable to remove indexes for key %s %s",
+            logger.warning("Unable to remove key %s indexes %s",
                            key, err.message)
         if fail:
             raise exceptions.FileNotFoundError("No such file %s" % key)
 
-    def s_read(self, path):
-        r = self._session.read_data(path, offset=0, size=0)
+    def s_read(self, path, offset=0, size=0):
+        r = self._session.read_data(path, offset=offset, size=size)
         r.wait()
         err = r.error()
         if err.code != 0:
@@ -162,14 +170,34 @@ class Storage(driver.Base):
         r.wait()
         err = r.error()
         if err.code != 0:
-            raise IOError("Writing failed {0}".format(err))
+            raise exceptions.UnspecifiedError("Writing failed %s" % err)
 
         # Set indexes
         r = self._session.update_indexes(key, list(tags), [key] * len(tags))
         r.wait()
         err = r.error()
         if err.code != 0:
-            raise IOError("Setting indexes failed {0}".format(err))
+            raise exceptions.UnspecifiedError("Indexe setting failed %s" % err)
+
+    def s_append(self, key, content):
+        session = self._session
+        session.ioflags = elliptics.io_flags.append
+
+        # set offset to resolve function overloading
+        r = session.write_data(key, content, offset=0)
+        r.wait()
+        err = r.error()
+        if err.code != 0:
+            raise exceptions.UnspecifiedError("Writing failed {0}".format(err))
+
+    def s_write_file(self, path, content):
+        tag, _, _ = path.rpartition('/')
+        if len(content) == 0:
+            content = "EMPTY"
+        logger.debug("put_content: write %s with tag %s", path, tag)
+        self.s_write(path, content, ('docker', tag))
+        self.create_fake_dir_struct(path)
+        return path
 
     @lru.get
     def get_content(self, path):
@@ -180,35 +208,75 @@ class Storage(driver.Base):
 
     @lru.set
     def put_content(self, path, content):
-        tag, _, _ = path.rpartition('/')
-        if len(content) == 0:
-            content = "EMPTY"
-        self.s_write(path, content, ('docker', tag))
-        spl_path = path.rsplit('/')[:-1]
-        while spl_path:
-            _path = '/'.join(spl_path)
-            _tag = '/'.join(spl_path[:-1])
-            spl_path.pop()
-            self.s_write(_path, "DIRECTORY", ('docker', _tag))
-        return path
+        logger.debug("put_content %s %d", path, len(content))
+        return self.s_write_file(path, content)
+
+    def create_fake_dir_struct(self, path):
+        """`path` is full filename (i.e. to create structure for file 'a/b/c'
+        `path` must be `a/b/c`).
+
+        To support listing and existance operations fake directory struct
+        is created on top of key-value and index operations.
+        Every file is tagged with his dirname index.
+        (i.e a/b/c. `c` content would be written as key `a/b/c` and marked
+        in index `a/b`
+        Fake directory file would be written as key `a/b` with a dummy content
+        to support existance operation.
+        Listing of `path` is performed as looking for all key marked
+        with `path` index.
+        """
+        logger.debug("creating fake directory structure %s", path)
+        # get parent dir for a given filepath
+        fakedir_key = os.path.dirname(path)
+        while True:
+            _tag = os.path.dirname(fakedir_key)
+            logger.debug("creating fake dir %s %s", fakedir_key, _tag)
+            self.s_write(fakedir_key, "DIRECTORY", ('docker', _tag))
+            fakedir_key = _tag
+            if not fakedir_key:  # root has been reached. fakedir_key is empty.
+                break
+        logger.debug("fake directory structure %s has been created", path)
 
     def stream_write(self, path, fp):
-        chunks = []
+        first_chunk = True
         while True:
             try:
                 buf = fp.read(self.buffer_size)
                 if not buf:
                     break
-                chunks += buf
-            except IOError:
+
+                # add buffer-in-the-middle
+                # not to write small chunks
+
+                if not first_chunk:
+                    self.s_append(path, buf)
+                else:
+                    # first of all the old file should be rewritten if exists.
+                    # all tags will be set up.
+                    self.s_write_file(path, buf)
+                    first_chunk = False
+
+            except IOError as err:
+                logger.error("unable to read from a given socket %s", err)
                 break
-        self.put_content(path, ''.join(chunks))
+        # should I clean not completely written file
+        # in case of error?
 
     def stream_read(self, path, bytes_range=None):
-        yield self.get_content(path)
+        logger.debug("read range %s from %s", str(bytes_range), path)
+        if not self.exists(path):
+            raise exceptions.FileNotFoundError(
+                'No such directory: \'{0}\''.format(path))
+
+        if bytes_range is None:
+            yield self.s_read(path)
+        else:
+            offset = bytes_range[0]
+            size = bytes_range[1] - bytes_range[0] + 1
+            yield self.s_read(path, offset=offset, size=size)
 
     def list_directory(self, path=None):
-        if path is None:
+        if path is None:  # pragma: no cover
             path = ""
 
         if not self.exists(path) and path:
@@ -219,18 +287,36 @@ class Storage(driver.Base):
             yield item
 
     def exists(self, path):
-        tag, _, _ = path.rpartition('/')
-        res = self.s_find(('docker', tag))
-        return path in res
+        logger.debug("Check existance of %s", path)
+        try:
+            # read is used instead of lookup
+            # just for future quorum reading check
+            self.s_read(path, 0, 1)
+        except exceptions.FileNotFoundError:
+            logger.debug("%s doesn't exist", path)
+            return False
+        else:
+            logger.debug("%s exists", path)
+            return True
 
     @lru.remove
     def remove(self, path):
         try:
             for subdir in self.list_directory(path):
                 self.s_remove(subdir)
-        except OSError as err:
+        except exceptions.FileNotFoundError as err:
             logger.warning(err)
         self.s_remove(path)
 
     def get_size(self, path):
-        return len(self.get_content(path))
+        logger.debug("get_size of %s", path)
+        r = self._session.lookup(path)
+        r.wait()
+        lookups = r.get()
+        err = r.error()
+        if err.code != 0:
+            raise exceptions.FileNotFoundError(
+                "Unable to get size of %s %s" % (path, err))
+        size = lookups[0].size
+        logger.debug("size of %s = %d", path, size)
+        return size
